#include <cfloat>
#include <cmath>
#include "Problem.h"
#include "Solution.h"
#include "flimflamSAControl.h"
#include "Random.h"

/*
  orOptTSP.cpp
  Program with the implementation of the function orOpt(), which performs
  the Or opt local search for the TSP, on a VRPSD instance.
  The initial solution given in input is modified  if a solution with
  smaller expected cost is found.
  The local optimum is with respect to the tsp length.*/

#include <fstream.h>
#include <iostream.h>
using namespace std;



inline bool SAcriterion(Random* rg, double delta, float Temp)
{
  //cout<<delta<<" "<<(exp((double)(-delta)/Temp))<<endl;
  return (rg->next() < (exp((double)(-delta)/Temp)));
}


void SAorOptTSP(Random* rnd, flimflamControl& control, Solution& initialSolution ){

  //Read problem pointer from input solution.
  Problem* pp = initialSolution.getProblem();

  float iniTemp = 0;
  int idleLength = 0, LengthIter = 0, iter=0;
  double worst_in_length = 0;
  double best_in_length = DBL_MAX;
  bool improved_in_length = false;
  float Temp;
  double delta = DBL_MAX;
  control.setSAOrOptParameters();
  
  //Initialize current best solution.
  Solution currSol(rnd, control, pp);
  currSol.copySolution(initialSolution);
  double globalBestFitness = currSol.computeExpectedCost();
  double currLen= currSol.computeTourLength();
  //cout << "currSol.expectedCost " << currSol.expectedCost << endl;//////
  //currSol.printOn(cout);/////	

  //Allocation for the best move.
  vector<int> bestMove(3,-1);
  
  
  //The lengths of the neighbour solution.
  double neighborCost;



#ifdef DEBUG
  currSol.printOn(cout);
  cout << "initial cost: " << currSol.expectedCost << endl;
#endif
  int neighborhoodSize = (pp->numberOfCustomers-3)
    + (pp->numberOfCustomers-2)
    + (pp->numberOfCustomers-1);
   
  //int TempLength = (int)control.rateTempLength;
  int TempLength = (int)(neighborhoodSize * control.rateTempLength);
 
  int stringLen = 3;
  int numSamp = 0;
  double recordDelta = 0;
  currSol.firstMove(stringLen);
  
  while (stringLen>=1)
    {
      do 
	{
	  recordDelta += fabs(currSol.computeMoveValueTSP(currLen) - currLen);
	  //cout<< currSol.computeMoveValue() - currFitness<<endl;
	  numSamp++;
	} while (currSol.nextMove(stringLen));
      stringLen--;
    }
  //cout<<(float)recordDelta/numSamp<<endl;
  iniTemp = control.initTempFactor*(float)recordDelta/numSamp;
  Temp = iniTemp;
  
  stringLen = 3;
  
  while(control.timeLeft() ){



 if (LengthIter > TempLength)
      {
	Temp = control.alfa * Temp;
	LengthIter = 0;
	if (improved_in_length)
	  idleLength=0;
	else
	  idleLength++;
	improved_in_length = false;
      }
    if (idleLength > control.idle_length_reheating)
      {
	Temp = Temp + iniTemp;
	idleLength = 0;  
	worst_in_length = 0;
	best_in_length = DBL_MAX;
      }    
    LengthIter++;
    iter++;

    #ifdef DEBUG 
    (*tempFile) <<iter<<"\t"<<Temp<<"\t"<<currFitness<<endl;
#endif
    //  cout << "firstMove("<<stringLen<< ") "<<endl;
    // << currSol.move[0] 
    // << " " << currSol.move[1] << " " << currSol.move[2]<< endl;//////////
    if (!currSol.nextMove(stringLen))
      {
	if (stringLen>1)
	  {
	    stringLen--;
	    currSol.firstMove(stringLen);
	  }
	else
	  {
	    stringLen=3;
	    currSol.firstMove(stringLen);
	  }
      }
    neighborCost = currSol.computeMoveValueTSP(currLen);
      //cout << "neighbCost " << neighborCost << endl;/////
   delta = neighborCost - currLen;
    if (delta <= 0)
      {
	currSol.setMove(currSol.getMove());
	  currSol.shiftTSP();
	  currLen= currSol.computeTourLength();
	  //	cout<<"+"<<endl;
	  if ( currSol.computeExpectedCost() < globalBestFitness ){
	    initialSolution.copySolution(currSol);
	    globalBestFitness = currSol.expectedCost;
	    control.setCurrentCost( globalBestFitness );
	    stringLen=3;
	  }
	  if (neighborCost > worst_in_length)
	    {
	      worst_in_length = neighborCost;
	      best_in_length = neighborCost;
	    }
	  if (neighborCost < best_in_length)
	    {
	      best_in_length = neighborCost;
	      improved_in_length = true;
	    }
      }
    else if (SAcriterion(rnd,delta,Temp))
	  {
	    //cout<<"accepted\n";
	    currSol.setMove(currSol.getMove());
	    currSol.shiftTSP();
	    stringLen=3;
	    if ( currSol.computeExpectedCost() < globalBestFitness ){
	      initialSolution.copySolution(currSol);
	      globalBestFitness = currSol.expectedCost;
	      control.setCurrentCost( globalBestFitness );
	      stringLen=3;
	    }
	  }
  }    //------------end main loop--------

#ifdef DEBUG
  delete tempFile;
#endif  
}

